import numpy as np
import pandas as pd

from pandas.testing import assert_series_equal
from numpy.testing import assert_allclose
import pytest
from datetime import timezone, timedelta

from pvlib import shading


@pytest.fixture
def test_system():
    syst = {
        "height": 1.0,
        "pitch": 2.0,
        "surface_tilt": 30.0,
        "surface_azimuth": 180.0,
        "rotation": -30.0,
    }  # rotation of right edge relative to horizontal
    syst["gcr"] = 1.0 / syst["pitch"]
    return syst


def test__ground_angle(test_system):
    ts = test_system
    x = np.array([0.0, 0.5, 1.0])
    angles = shading.ground_angle(ts["surface_tilt"], ts["gcr"], x)
    expected_angles = np.array([0.0, 5.866738789543952, 9.896090638982903])
    assert np.allclose(angles, expected_angles)


def test__ground_angle_zero_gcr():
    surface_tilt = 30.0
    x = np.array([0.0, 0.5, 1.0])
    angles = shading.ground_angle(surface_tilt, 0, x)
    expected_angles = np.array([0, 0, 0])
    assert np.allclose(angles, expected_angles)


@pytest.fixture
def surface_tilt():
    idx = pd.date_range("2019-01-01", freq="h", periods=3)
    return pd.Series([0, 20, 90], index=idx)


@pytest.fixture
def masking_angle(surface_tilt):
    # masking angles for the surface_tilt fixture,
    # assuming GCR=0.5 and height=0.25
    return pd.Series([0.0, 11.20223712, 20.55604522], index=surface_tilt.index)


@pytest.fixture
def average_masking_angle(surface_tilt):
    # average masking angles for the surface_tilt fixture, assuming GCR=0.5
    return pd.Series([0.0, 7.20980655, 13.779867461], index=surface_tilt.index)


@pytest.fixture
def shading_loss(surface_tilt):
    # diffuse shading loss values for the average_masking_angle fixture
    return pd.Series([0, 0.00395338, 0.01439098], index=surface_tilt.index)


def test_masking_angle_series(surface_tilt, masking_angle):
    # series inputs and outputs
    masking_angle_actual = shading.masking_angle(surface_tilt, 0.5, 0.25)
    assert_series_equal(masking_angle_actual, masking_angle)


def test_masking_angle_scalar(surface_tilt, masking_angle):
    # scalar inputs and outputs, including zero
    for tilt, angle in zip(surface_tilt, masking_angle):
        masking_angle_actual = shading.masking_angle(tilt, 0.5, 0.25)
        assert np.isclose(masking_angle_actual, angle)


def test_masking_angle_zero_gcr(surface_tilt):
    # scalar inputs and outputs, including zero
    for tilt in surface_tilt:
        masking_angle_actual = shading.masking_angle(tilt, 0, 0.25)
        assert np.isclose(masking_angle_actual, 0)


def test_masking_angle_passias_series(surface_tilt, average_masking_angle):
    # pandas series inputs and outputs
    masking_angle_actual = shading.masking_angle_passias(surface_tilt, 0.5)
    assert_series_equal(masking_angle_actual, average_masking_angle)


def test_masking_angle_passias_scalar(surface_tilt, average_masking_angle):
    # scalar inputs and outputs, including zero
    for tilt, angle in zip(surface_tilt, average_masking_angle):
        masking_angle_actual = shading.masking_angle_passias(tilt, 0.5)
        assert np.isclose(masking_angle_actual, angle)


def test_sky_diffuse_passias_series(average_masking_angle, shading_loss):
    # pandas series inputs and outputs
    actual_loss = shading.sky_diffuse_passias(average_masking_angle)
    assert_series_equal(shading_loss, actual_loss)


def test_sky_diffuse_passias_scalar(average_masking_angle, shading_loss):
    # scalar inputs and outputs
    for angle, loss in zip(average_masking_angle, shading_loss):
        actual_loss = shading.sky_diffuse_passias(angle)
        assert np.isclose(loss, actual_loss)


@pytest.fixture
def true_tracking_angle_and_inputs_NREL():
    # data from NREL 'Slope-Aware Backtracking for Single-Axis Trackers'
    # doi.org/10.2172/1660126 ; Accessed on 2023-11-06.
    tzinfo = timezone(timedelta(hours=-5))
    axis_tilt_angle = 9.666  # deg
    axis_azimuth_angle = 195.0  # deg
    timedata = pd.DataFrame(
        columns=("Apparent Elevation", "Solar Azimuth", "True-Tracking"),
        data=(
            (2.404287, 122.791770, -84.440),
            (11.263058, 133.288729, -72.604),
            (18.733558, 145.285552, -59.861),
            (24.109076, 158.939435, -45.578),
            (26.810735, 173.931802, -28.764),
            (26.482495, 189.371536, -8.475),
            (23.170447, 204.136810, 15.120),
            (17.296785, 217.446538, 39.562),
            (9.461862, 229.102218, 61.587),
            (0.524817, 239.330401, 79.530),
        ),
    )
    timedata.index = pd.date_range(
        "2019-01-01T08", "2019-01-01T17", freq="1h", tz=tzinfo
    )
    timedata["Apparent Zenith"] = 90.0 - timedata["Apparent Elevation"]
    return (axis_tilt_angle, axis_azimuth_angle, timedata)


@pytest.fixture
def projected_solar_zenith_angle_edge_cases():
    premises_and_result_matrix = pd.DataFrame(
        data=[
            # s_zen | s_azm | ax_tilt | ax_azm | psza
            [   0,       0,      0,        0,      0],
            [   0,     180,      0,        0,      0],
            [   0,       0,      0,      180,      0],
            [   0,     180,      0,      180,      0],
            [  45,       0,      0,      180,      0],
            [  45,      90,      0,      180,    -45],
            [  45,     270,      0,      180,     45],
            [  45,      90,     90,      180,    -90],
            [  45,     270,     90,      180,     90],
            [  45,      90,     90,        0,     90],
            [  45,     270,     90,        0,    -90],
            [  45,      45,     90,      180,   -135],
            [  45,     315,     90,      180,    135],
        ],
        columns=["solar_zenith", "solar_azimuth", "axis_tilt", "axis_azimuth",
                 "psza"],
    )
    return premises_and_result_matrix


def test_projected_solar_zenith_angle_numeric(
    true_tracking_angle_and_inputs_NREL,
    projected_solar_zenith_angle_edge_cases
):
    psza_func = shading.projected_solar_zenith_angle
    axis_tilt, axis_azimuth, timedata = true_tracking_angle_and_inputs_NREL
    # test against data provided by NREL
    psz = psza_func(
        timedata["Apparent Zenith"],
        timedata["Solar Azimuth"],
        axis_tilt,
        axis_azimuth,
    )
    assert_allclose(psz, timedata["True-Tracking"], atol=1e-3)
    # test by changing axis azimuth and tilt
    psza = psza_func(
        timedata["Apparent Zenith"],
        timedata["Solar Azimuth"],
        -axis_tilt,
        axis_azimuth - 180,
    )
    assert_allclose(psza, -timedata["True-Tracking"], atol=1e-3)

    # test edge cases
    solar_zenith, solar_azimuth, axis_tilt, axis_azimuth, psza_expected = (
        v for _, v in projected_solar_zenith_angle_edge_cases.items()
    )
    psza = psza_func(
        solar_zenith,
        solar_azimuth,
        axis_tilt,
        axis_azimuth,
    )
    assert_allclose(psza, psza_expected, atol=1e-9)


@pytest.mark.parametrize(
    "cast_type, cast_func",
    [
        (float, lambda x: float(x)),
        (np.ndarray, lambda x: np.array([x])),
        (pd.Series, lambda x: pd.Series(data=[x])),
    ],
)
def test_projected_solar_zenith_angle_datatypes(
    cast_type, cast_func, true_tracking_angle_and_inputs_NREL
):
    psz_func = shading.projected_solar_zenith_angle
    axis_tilt, axis_azimuth, timedata = true_tracking_angle_and_inputs_NREL
    sun_apparent_zenith = timedata["Apparent Zenith"].iloc[0]
    sun_azimuth = timedata["Solar Azimuth"].iloc[0]

    axis_tilt, axis_azimuth, sun_apparent_zenith, sun_azimuth = (
        cast_func(sun_apparent_zenith),
        cast_func(sun_azimuth),
        cast_func(axis_tilt),
        cast_func(axis_azimuth),
    )
    psz = psz_func(sun_apparent_zenith, axis_azimuth, axis_tilt, axis_azimuth)
    assert isinstance(psz, cast_type)
